/*
 *  mdaOverdriveProcessor.cpp
 *  mda-vst3
 *
 *  Created by Arne Scheffler on 6/14/08.
 *
 *  mda VST Plug-ins
 *
 *  Copyright (c) 2008 Paul Kellett
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#include "mdaOverdriveProcessor.h"
#include "mdaOverdriveController.h"

#include <math.h>
#include <time.h>

namespace Steinberg {
	namespace Vst {
		namespace mda {

			//-----------------------------------------------------------------------------
			FUID OverdriveProcessor::uid(0x12051428, 0x3400291E, 0x01002916, 0x05127E92);

			//-----------------------------------------------------------------------------
			OverdriveProcessor::OverdriveProcessor()
			{
				for (int i = 0; i < MAXSQRT; i++)
				{
					aSqrt[i] = (float)sqrt((1.0f*(i)) / (1.0f*MAXSQRT));
				}
				setControllerClass(OverdriveController::uid);
				allocParameters(OverdriveController::kNumParams);
			}

			//-----------------------------------------------------------------------------
			OverdriveProcessor::~OverdriveProcessor()
			{
			}

			//-----------------------------------------------------------------------------
			tresult PLUGIN_API OverdriveProcessor::initialize(FUnknown* context)
			{
				tresult res = BaseProcessor::initialize(context);
				if (res == kResultTrue)
				{
					addAudioInput(USTRING("Stereo In"), SpeakerArr::kStereo);
					addAudioOutput(USTRING("Stereo Out"), SpeakerArr::kStereo);

					params[OverdriveController::kParamDrive] = OverdriveController::kParamDefDrive;
					params[OverdriveController::kParamFilterCoeff] = OverdriveController::kParamDefFilterCoeff;
					params[OverdriveController::kParamInput] = OverdriveController::kParamDefInput;
					params[OverdriveController::kParamMix] = OverdriveController::kParamDefMix;
					params[OverdriveController::kParamMuffle] = OverdriveController::kParamDefMuffle;
					params[OverdriveController::kParamOutput] = OverdriveController::kParamDefOutput;
					params[OverdriveController::kParamDriveExpo] = OverdriveController::kParamDefDriveExpo;
					params[OverdriveController::kParamDriveMode] = OverdriveController::kParamDefDriveMode;
					params[OverdriveController::kParamDriveModeCoeff] = OverdriveController::kParamDefDriveModeCoeff;
					params[OverdriveController::kParamAberration1] = OverdriveController::kParamDefAberration1;
					params[OverdriveController::kParamAberration2] = OverdriveController::kParamDefAberration2;

					srand(time(NULL));

					recalculate();
				}
				return res;
			}

			//-----------------------------------------------------------------------------
			tresult PLUGIN_API OverdriveProcessor::terminate()
			{
				return BaseProcessor::terminate();
			}

			//-----------------------------------------------------------------------------
			tresult PLUGIN_API OverdriveProcessor::setActive(TBool state)
			{
				return BaseProcessor::setActive(state);
			}

			//-----------------------------------------------------------------------------
			void OverdriveProcessor::doProcessing(ProcessData& data)
			{
				int32 sampleFrames = data.numSamples;

				float* in1 = data.inputs[0].channelBuffers32[0];
				float* in2 = data.inputs[0].channelBuffers32[1];
				float* out1 = data.outputs[0].channelBuffers32[0];
				float* out2 = data.outputs[0].channelBuffers32[1];

				float a, b, c, d;
				float i = drive, g = gain, aa, bb, dmc = dmodecoeff;
				bool m = modeSqrt;
				float f = filt, fa = filt1, fb = filt2;
				int32 s;// , arrb1, arrb2, r1, r2;
				float r, r12, ar, br; /*=ar12*/

				--in1;
				--in2;
				--out1;
				--out2;
				while (--sampleFrames >= 0)
				{
					a = *++in1 * in;
					b = *++in2 * in;

					//overdrive
					aa = (a > 0.0f) ? (float)sqrt(a) : (float)-sqrt(-a);
					bb = (b > 0.0f) ? (float)sqrt(b) : (float)-sqrt(-b);
					if (!m) {
						for (int q = 0; q < sqrtDuration; q++) {
							s = abs((int32)(MAXSQRT*aa));
							if (s >= MAXSQRT)
								s = MAXSQRT - 1;
							aa = (aa > 0.0f) ? aSqrt[s] : -aSqrt[s];
							s = abs((int32)(MAXSQRT*bb));
							if (s >= MAXSQRT)
								s = MAXSQRT - 1;
							bb = (bb > 0.0f) ? aSqrt[s] : -aSqrt[s];
						}
					}

					// aberration + -
					//arrb1 = (int32)(aberration1);
					//r1 = 0;
					//if (arrb1>0)
					//	r1 = rand() % arrb1 + 1; // Randomizing the number between 1-arrb1.
					r = 1.0f*rand() / (1.0f *  RAND_MAX);

					r12 = (aberration1 + aberration2) * r;
					ar = aa*(1.0f - aberration2 + r12);
					//aa = (aa > 0.0f) ? aa += ar : aa -= ar;
					aa += ar;
					br = bb * (1.0f - aberration2 + r12);
					//bb = (bb > 0.0f) ? bb += br : bb -= br;
					bb += br;

					fa = fa + f * (i * (aa - a) + a - fa);                //filter 
					fb = fb + f * (i * (bb - b) + b - fb);

					c = fa;;
					d = fb;;

					*++out1 = (c*mix + a*(1 - mix)) * g;
					*++out2 = (d*mix + b*(1 - mix)) * g;
				}
				if (fabs(fa) > 1.0e-10) filt1 = fa; else filt1 = 0.0f; //catch denormals
				if (fabs(fb) > 1.0e-10) filt2 = fb; else filt2 = 0.0f;
			}

			//-----------------------------------------------------------------------------
			void OverdriveProcessor::recalculate()
			{
				dmodecoeff = 10.0f * params[OverdriveController::kParamDriveModeCoeff];
				sqrtDuration = (int32)dmodecoeff + 1;

				modeSqrt = (params[OverdriveController::kParamDriveMode] < 0.5f);
				if (params[OverdriveController::kParamDriveExpo] < 0.5f)
					drive = params[OverdriveController::kParamDrive];
				else
					drive = (float)pow(10.0f, (float)(2.0f * params[OverdriveController::kParamDrive] - 1.0f)); ;
				mix = params[OverdriveController::kParamMix];
				in = (float)pow(10.0f, (float)(2.0f * params[OverdriveController::kParamInput] - 1.0f));
				filterCoeff = 10.0f*(params[OverdriveController::kParamFilterCoeff] - 1.0f);
				//filterCoeff = -1.6f;
				filt = (float)pow(10.0f, filterCoeff * params[OverdriveController::kParamMuffle]);
				gain = (float)pow(10.0f, (float)(2.0f * params[OverdriveController::kParamOutput] - 1.0f));

				aberration1 = 100.f * params[OverdriveController::kParamAberration1];
				aberration2 = 100.f * params[OverdriveController::kParamAberration2];

				//float r = (1.0f * rand()) / (1.0f * RAND_MAX);
				//ar12 = (aberration1 + aberration2) * r * 10.0f;
			}

		}
	}
} // namespaces

